! string.f90 - ISO F95 cpmpatible string module for Fortran90
! Copyright (C) TOYODA Eizi, 2000.  All rights reserved.
! vi: set ts=8 sw=4:

! 
!	{W[ ISO/IEC 1539-2 : 1999 ŋK肳ĂA
!	Fortran 95 ̂߂̉ϒ񃂃W[ɑ̂
!	Fortran 90 Ŏ̂łB
!	@\B
!
!	ߎSȂƂɁAFortran 90 ɂ͎ϐ̎JȂ̂ŁA
!	\̂̂߂Ɋt͎蓮ŊJȂ
!	Ă܂B
!	̂߂̎葱Ă܂ ISO ƌ݊ł
!	ȂȂ邽߁AȂ񂾂ςł͂邯ǁA{Cuł͒[
!	Zbg邱ƂpƂ݂ȂƂɂB

#ifndef HAVE_ELEMENTAL
! remove keyword 'elemental'
#define	elemental
#endif

module iso_varying_string

!    implicit none

    private
    public varying_string, len, var_str, char, get, put, put_line, &
	& assignment(=), operator(==), operator(/=), operator(//), &
	& operator(<), operator(<=), operator(>), operator(>=)

    type varying_string
    	private
	integer::				len = 0
	character, dimension(:), pointer::	buffer
    end type

    interface len
	module procedure string_len
    end interface

    interface var_str
	module procedure char_to_string
    end interface

    interface char
	module procedure string_to_char_all
	module procedure string_to_char_length
    end interface

    interface get
	module procedure string_get
	module procedure string_get_default
    end interface

    interface put
	module procedure string_put
	module procedure string_put_default
	module procedure char_put
	module procedure char_put_default
    end interface

    interface put_line
	module procedure string_put_line
	module procedure string_put_line_default
	module procedure char_put_line
	module procedure char_put_line_default
    end interface

    interface scan
	module procedure string_scan_string
	module procedure string_scan_char
    end interface

    interface verify
	module procedure string_verify_string
	module procedure string_verify_char
    end interface

    interface index
	module procedure string_index_string
	module procedure string_index_char
	module procedure char_index_string
    end interface

    interface assignment(=)
	module procedure string_let_char
	module procedure char_let_string
	module procedure string_let_string
    end interface

    interface operator(//)
	module procedure string_add_string
	module procedure char_add_string
	module procedure string_add_char
    end interface

    interface operator(==)
	module procedure string_eq_string
	module procedure string_eq_char
	module procedure char_eq_string
    end interface

    interface operator(/=)
	module procedure string_ne_string
	module procedure string_ne_char
	module procedure char_ne_string
    end interface

    interface operator(<)
	module procedure string_lt_string
	module procedure string_lt_char
	module procedure char_lt_string
    end interface

    interface operator(<=)
	module procedure string_le_string
	module procedure string_le_char
	module procedure char_le_string
    end interface

    interface operator(>)
	module procedure string_gt_string
	module procedure string_gt_char
	module procedure char_gt_string
    end interface

    interface operator(>=)
	module procedure string_ge_string
	module procedure string_ge_char
	module procedure char_ge_string
    end interface

    !
    ! rẐ߂̃V{
    !

    integer, parameter::	CMP_LESSTHAN = -1
    integer, parameter::	CMP_EQUIVALENT = 0
    integer, parameter::	CMP_GREATERTHAN = 1
    integer, parameter::	CMP_ERROR = -100

contains

    !
    !  len ֐̎
    !

    elemental integer function string_len(string) result(result)
	type(varying_string), intent(in)::	string
    continue
    	result = string%len
    end function

    !
    !  var_str ֐̎
    !

    type(varying_string) function char_to_string(char) result(result)
	character(len=*), intent(in)::	char
    continue
	result = char
    end function

    !
    !  char ֐̎
    !

    function string_to_char_all(string) result(result)
	type(varying_string), intent(in)::	string
	character(len = string%len)::		result
    continue
	result = string
    end function

    function string_to_char_length(string, length) result(result)
	type(varying_string), intent(in)::	string
	integer, intent(in)::			length
	character(len = length)::		result
    continue
	result = string
    end function

    !
    ! ̑̎
    !

    subroutine string_let_string(lhs, rhs)
	type(varying_string), intent(out)::	lhs
	type(varying_string), intent(in)::	rhs
    continue
	lhs = char(rhs)
    end subroutine

    subroutine string_let_char(string, chars)
	type(varying_string), intent(out)::	string
	character(len=*), intent(in)::		chars
	integer::				length, i
    continue
	length = len(chars)
	call string_allocate(string, length)
	if (length == 0) return
	do, i = 1,  length
	    string%buffer(i) = chars(i:i)
	enddo
    end subroutine

    subroutine char_let_string(char, string)
	character(len=*), intent(out)::		char
	type(varying_string), intent(in)::	string
	integer::				length
    continue
	length = min(len(char), len(string))
	if (length == 0) return
	char = transfer(string%buffer(1: length), char(1:length))
    end subroutine

    !
    !  // Zq̎
    !

    function string_add_string(lhs, rhs) result(result)
	type(varying_string)::			result
	type(varying_string), intent(in)::	lhs, rhs
	integer::				lhslen, rhslen
    continue
	lhslen = len(lhs)
	rhslen = len(rhs)
	call string_allocate(result, lhslen + rhslen)
	if (lhslen > 0) then
	    result%buffer(1: lhslen) = lhs%buffer(:)
	endif
	if (rhslen > 0) then
	    result%buffer(lhslen+1: lhslen+rhslen) = rhs%buffer(:)
	endif
    end function

    function char_add_string(char, string) result(result)
	type(varying_string)::			result
	character(len=*), intent(in)::		char
	type(varying_string), intent(in)::	string
    continue
	result = string_add_string(var_str(char), string)
    end function

    function string_add_char(string, char) result(result)
	type(varying_string)::			result
	type(varying_string), intent(in)::	string
	character(len=*), intent(in)::		char
    continue
	result = string_add_string(string, var_str(char))
    end function

    !
    ! ̉Zq == ̎
    !

    logical function string_eq_string(lhs, rhs) result(result)
	type(varying_string), intent(in)::	lhs, rhs
    continue
	result = string_cmp_string(lhs, rhs) == CMP_EQUIVALENT
    end function

    logical function string_eq_char(string, char) result(result)
	type(varying_string), intent(in)::	string
	character(len=*), intent(in)::		char
    continue
	result = string_eq_string(string, var_str(char))
    end function

    logical function char_eq_string(char, string) result(result)
	character(len=*), intent(in)::		char
	type(varying_string), intent(in)::	string
    continue
	result = string_eq_string(var_str(char), string)
    end function

    !
    ! ̉Zq /= ̎
    !

    logical function string_ne_string(lhs, rhs) result(result)
	type(varying_string), intent(in)::	lhs, rhs
    continue
	result = .not. string_eq_string(lhs, rhs)
    end function

    logical function string_ne_char(string, char) result(result)
	type(varying_string), intent(in)::	string
	character(len=*), intent(in)::		char
    continue
	result = .not. string_eq_char(string, char)
    end function

    logical function char_ne_string(char, string) result(result)
	character(len=*), intent(in)::		char
	type(varying_string), intent(in)::	string
    continue
	result = .not. char_eq_string(char, string)
    end function

    !
    ! ̉Zq < ̎
    !

    logical function string_lt_string(lhs, rhs) result(result)
	type(varying_string), intent(in)::	lhs, rhs
    continue
	result = string_cmp_string(lhs, rhs) == CMP_LESSTHAN
    end function

    logical function string_lt_char(string, char) result(result)
	type(varying_string), intent(in)::	string
	character(len=*), intent(in)::		char
    continue
	result = string_lt_string(string, var_str(char))
    end function

    logical function char_lt_string(char, string) result(result)
	character(len=*), intent(in)::		char
	type(varying_string), intent(in)::	string
    continue
	result = string_lt_string(var_str(char), string)
    end function

    !
    ! ̉Zq <= ̎
    !

    logical function string_le_string(lhs, rhs) result(result)
	type(varying_string), intent(in)::	lhs, rhs
	integer::				cmp
    continue
	cmp = string_cmp_string(lhs, rhs) 
	result = (cmp == CMP_LESSTHAN) .or. (cmp == CMP_EQUIVALENT)
    end function

    logical function string_le_char(string, char) result(result)
	type(varying_string), intent(in)::	string
	character(len=*), intent(in)::		char
    continue
	result = string_le_string(string, var_str(char))
    end function

    logical function char_le_string(char, string) result(result)
	character(len=*), intent(in)::		char
	type(varying_string), intent(in)::	string
    continue
	result = string_le_string(var_str(char), string)
    end function

    !
    ! ̉Zq > ̎
    !

    logical function string_gt_string(lhs, rhs) result(result)
	type(varying_string), intent(in)::	lhs, rhs
    continue
	result = string_cmp_string(lhs, rhs) == CMP_GREATERTHAN
    end function

    logical function string_gt_char(string, char) result(result)
	type(varying_string), intent(in)::	string
	character(len=*), intent(in)::		char
    continue
	result = string_gt_string(string, var_str(char))
    end function

    logical function char_gt_string(char, string) result(result)
	character(len=*), intent(in)::		char
	type(varying_string), intent(in)::	string
    continue
	result = string_gt_string(var_str(char), string)
    end function

    !
    ! ̉Zq >= ̎
    !

    logical function string_ge_string(lhs, rhs) result(result)
	type(varying_string), intent(in)::	lhs, rhs
	integer::				cmp
    continue
	cmp = string_cmp_string(lhs, rhs) 
	result = (cmp == CMP_GREATERTHAN) .or. (cmp == CMP_EQUIVALENT)
    end function

    logical function string_ge_char(string, char) result(result)
	type(varying_string), intent(in)::	string
	character(len=*), intent(in)::		char
    continue
	result = string_ge_string(string, var_str(char))
    end function

    logical function char_ge_string(char, string) result(result)
	character(len=*), intent(in)::		char
	type(varying_string), intent(in)::	string
    continue
	result = string_ge_string(var_str(char), string)
    end function

    !
    ! o
    !

    subroutine string_get_default(string, maxlen, iostat)
	type(varying_string), intent(out)::	string
	integer, intent(in), optional::		maxlen
	integer, intent(out), optional::	iostat
    continue
	call string_get(-1, string, maxlen, iostat)
    end subroutine

    subroutine string_get(unit, string, maxlen, iostat)
	integer, intent(in)::			unit
	type(varying_string), intent(out)::	string
	integer, intent(in), optional::		maxlen
	integer, intent(out), optional::	iostat
	integer::		alreadyread, buflen, nowread, ios, maxsize
	integer, parameter::	BUFFERSIZE = 80
	character(len = BUFFERSIZE)::		buffer
    continue
	if (present(maxlen)) then
	    maxsize = maxlen
	else
	    maxsize = huge(maxsize)
	endif
	alreadyread = 0
	string = ''
	do
	    if (alreadyread >= maxsize) return
	    buflen = min(BUFFERSIZE, maxsize - alreadyread)
	    if (unit >= 0) then
		read(unit=unit, fmt='(A)', advance='NO', &
		    & size=nowread, eor=100, iostat=ios) buffer(1: buflen)
	    else
		read(unit=*, fmt='(A)', advance='NO', &
		    & size=nowread, eor=100, iostat=ios) buffer(1: buflen)
	    endif
	    if (ios /= 0) then
		if (present(iostat)) then
		    iostat = ios
		    return
		else
		    print *, 'get_string: read error ', ios
		    stop
		endif
	    endif
	    alreadyread = alreadyread + nowread
	    string = string // buffer(1: nowread)
	enddo
	if (present(iostat)) iostat = 0
	return
	! in case of EOR
100	string = string // buffer(1: nowread)
	if (present(iostat)) iostat = 0
    end subroutine string_get

    subroutine char_put_default(char, iostat)
	character(len=*), intent(in)::		char
	integer, intent(out), optional::	iostat
    continue
	call char_put(-1, char, iostat)
    end subroutine

    subroutine char_put(unit, char, iostat)
	integer, intent(in)::			unit
	character(len=*), intent(in)::		char
	integer, intent(out), optional::	iostat
	integer:: ios
    continue
	if (unit >= 0) then
	    write(unit=unit, fmt='(A)', advance='NO', iostat=ios) char
	else
	    write(unit=*, fmt='(A)', advance='NO', iostat=ios) char
	endif
	if (present(iostat)) then
	    iostat = ios
	else
	    if (ios /= 0) then
	        print *, 'char_put: write error ', ios
	    endif
	endif
    end subroutine

    subroutine string_put_default(string, iostat)
	type(varying_string), intent(in)::	string
	integer, intent(out), optional::	iostat
    continue
	call string_put(-1, string, iostat)
    end subroutine

    subroutine string_put(unit, string, iostat)
	integer, intent(in)::			unit
	type(varying_string), intent(in)::	string
	integer, intent(out), optional::	iostat
	integer:: ios
    continue
	if (unit >= 0) then
	    write(unit=unit, fmt='(A)', advance='NO', iostat=ios) char(string)
	else
	    write(unit=*, fmt='(A)', advance='NO', iostat=ios) char(string)
	endif
	if (present(iostat)) then
	    iostat = ios
	else
	    if (ios /= 0) then
	        print *, 'string_put: write error ', ios
	    endif
	endif
    end subroutine

    subroutine char_put_line_default(char, iostat)
	character(len=*), intent(in)::		char
	integer, intent(out), optional::	iostat
    continue
	call char_put_line(-1, char, iostat)
    end subroutine

    subroutine char_put_line(unit, char, iostat)
	integer, intent(in)::			unit
	character(len=*), intent(in)::		char
	integer, intent(out), optional::	iostat
	integer:: ios
    continue
	if (unit >= 0) then
	    write(unit=unit, fmt='(A)', advance='YES', iostat=ios) char
	else
	    write(unit=*, fmt='(A)', advance='YES', iostat=ios) char
	endif
	if (present(iostat)) then
	    iostat = ios
	else
	    if (ios /= 0) then
		print *, 'char_put_line: write error ', ios
	    endif
	endif
    end subroutine

    subroutine string_put_line_default(string, iostat)
	type(varying_string), intent(in)::	string
	integer, intent(out), optional::	iostat
    continue
	call string_put_line(-1, string, iostat)
    end subroutine

    subroutine string_put_line(unit, string, iostat)
	integer, intent(in)::			unit
	type(varying_string), intent(in)::	string
	integer, intent(out), optional::	iostat
	integer:: ios
    continue
	if (unit >= 0) then
	    write(unit=unit, fmt='(A)', advance='YES', iostat=ios) char(string)
	else
	    write(unit=*, fmt='(A)', advance='YES', iostat=ios) char(string)
	endif
	if (present(iostat)) then
	    iostat = ios
	else
	    if (ios /= 0) then
		print *, 'string_put_line: write error ', ios
	    endif
	endif
    end subroutine

    !
    ! --- gݍ݊֐̏㏑ ---
    !

    !
    ! index ̑p
    !

    function string_index_string(string, substring, back) result(result)
	integer::				result
	type(varying_string), intent(in)::	string, substring
	logical, optional::			back
    continue
	result = index(char(string), char(substring), back)
    end function

    function string_index_char(string, substring, back) result(result)
	integer::				result
	type(varying_string), intent(in)::	string
	character(len = *), intent(in)::	substring
	logical, optional::			back
    continue
	result = index(char(string), substring, back)
    end function

    function char_index_string(string, substring, back) result(result)
	integer::				result
	character(len = *), intent(in)::	string
	type(varying_string), intent(in)::	substring
	logical, optional::			back
    continue
	result = index(string, char(substring), back)
    end function

    !
    ! scan ̑p
    !

    function string_scan_string(string, set, back) result(result)
	integer::				result
	type(varying_string), intent(in)::	string, set
	logical, optional::			back
    continue
	result = scan(char(string), char(set), back)
    end function

    function string_scan_char(string, set, back) result(result)
	integer::				result
	type(varying_string), intent(in)::	string
	character(len = *), intent(in)::	set
	logical, optional::			back
    continue
	result = scan(char(string), set, back)
    end function

    !
    ! verify ̑p
    !

    function string_verify_string(string, set, back) result(result)
	integer::				result
	type(varying_string), intent(in)::	string, set
	logical, optional::			back
    continue
	result = verify(char(string), char(set), back)
    end function

    function string_verify_char(string, set, back) result(result)
	integer::				result
	type(varying_string), intent(in)::	string
	character(len = *), intent(in)::	set
	logical, optional::			back
    continue
	result = verify(char(string), set, back)
    end function

    !
    ! --- Iɗp ---
    !

    ! rZq
    integer function string_cmp_string(lhs, rhs) result(result)
	type(varying_string), intent(in)::	lhs, rhs
	integer::				lhs_len, rhs_len, i
	character::				lhs_char, rhs_char
    continue
	if (.not. associated(lhs%buffer) &
	& .or. .not. associated(rhs%buffer)) then
	    result = CMP_ERROR
	    return
	endif
	result = CMP_EQUIVALENT
	lhs_len = len(lhs)
	rhs_len = len(rhs)
	do, i = 1, max(lhs_len, len(rhs))
	    if (i <= lhs_len) then
		lhs_char = lhs%buffer(i)
	    else
		lhs_char = ' '
	    endif
	    if (i <= rhs_len) then
		rhs_char = rhs%buffer(i)
	    else
		rhs_char = ' '
	    endif
	    if (lhs_char > rhs_char) then
		result = CMP_GREATERTHAN
		return
	    else if (lhs_char < rhs_char) then
		result = CMP_LESSTHAN
		return
	    endif
	enddo
    end function

    ! ̈p (Fortran 95 ȍ~ł͕sv)
    subroutine string_dispose(string)
	type(varying_string), intent(inout)::	string
    continue
	if (.not. associated(string%buffer)) return
	deallocate(string%buffer)
	nullify(string%buffer)
	string%len = 0
    end subroutine

    ! length ̗̈t
    subroutine string_allocate(string, length)
	type(varying_string), intent(inout)::	string
	integer, intent(in)::			length
    continue
	if (length <= 0) then
	    call string_dispose(string)
	    string%len =0
	    return
	endif
	if (associated(string%buffer)) deallocate(string%buffer)
	allocate(string%buffer(1:length))
	string%len = length
    end subroutine

end module iso_varying_string
